# Copyright 2020 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import sys
from collections.abc import Sequence
from gettext import gettext as _

import gi

try:
    gi.require_version('Gdk', '4.0')
    gi.require_version('Gtk', '4.0')
    gi.require_version('Adw', '1')
    gi.require_version('GtkSource', '5')
    gi.require_version('Soup', '3.0')
    from gi.repository import Adw, Gio, GLib
except ImportError or ValueError as exc:
    print('Error: GIR dependencies not met.', exc)
    exit()

try:
    import fontTools  # noqa
    import brotli  # noqa
except ModuleNotFoundError as exc:
    print('Error: Python dependencies not met.', exc)
    exit()

from webfontkitgenerator.window import Window


class Application(Adw.Application):
    def __init__(self, version: str):
        super().__init__(
            application_id='com.rafaelmardojai.WebfontKitGenerator',
            flags=Gio.ApplicationFlags.HANDLES_OPEN,
        )
        GLib.set_application_name(_('Webfont Bundler'))

        self.version = version
        self.window: Window | None = None

        self.setup_actions()

    def do_startup(self):
        Adw.Application.do_startup(self)

    def do_activate(self):
        self.window = self.props.active_window  # type: ignore
        if not self.window:
            self.window = Window(application=self)
        self.window.present()

    def do_open(self, files: Sequence[Gio.File], _n_files, _hint):
        # Activate window first
        self.activate()
        # Load fonts
        if self.window:
            self.window.load_fonts(files)

    def setup_actions(self):
        about = Gio.SimpleAction.new('about', None)
        about.connect('activate', self._on_about)
        self.add_action(about)

        self.set_accels_for_action('win.open', ['<Ctl>o'])
        self.set_accels_for_action('win.google', ['<Ctl>g'])
        self.set_accels_for_action('win.set-dir', ['<Ctl>d'])
        self.set_accels_for_action('win.generate', ['<Ctl>Return'])
        self.set_accels_for_action('win.back', ['Escape'])

    def _on_about(self, _action, _param):
        dialog = Adw.AboutDialog.new_from_appdata(
            '/com/rafaelmardojai/WebfontKitGenerator/appdata.xml',
            str(self.version),
        )
        dialog.props.translator_credits = _("translator-credits")
        dialog.present(self.window)


def main(version):
    app = Application(version)
    return app.run(sys.argv)



# Copyright 2020 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import os
import typing
from collections.abc import Sequence
from gettext import gettext as _
from urllib.parse import unquote

from fontTools.ttLib import TTFont
from gi.repository import Adw, Gio, GLib

from webfontkitgenerator.font import Font, FontData

if typing.TYPE_CHECKING:
    from webfontkitgenerator.window import Window

WEIGHTS = {
    'thin': '100',
    'extralight': '200',
    'light': '300',
    'regular': '400',
    'medium': '500',
    'semibold': '600',
    'bold': '700',
    'extrabold': '800',
    'black': '900',
}

WIDTHS = {
    'ultracondensed': '50%',
    'extracondensed': '62.5%',
    'condensed': '75%',
    'semicondensed': '87.5%',
    'normal': '100%',
    'semiexpanded': '112.5%',
    'expanded': '125%',
    'extraexpanded': '150%',
    'ultraexpanded': '200%',
}


class Loader(object):
    def __init__(self, window: Window, model: Gio.ListStore):
        self.window = window
        self.model = model

    def load(self, files: Sequence[Gio.File]):
        self.window.processing = True

        for f in files:
            try:
                path = f.get_path()
                if path:
                    path = unquote(path)

                    if os.path.exists(path) and os.access(path, os.R_OK):
                        ttfont = TTFont(path, lazy=True)
                        data = self._get_font_data(ttfont)
                        ttfont.close()
                        font = Font(path, data)
                        GLib.idle_add(self.model.append, font)
                    else:
                        error_text = _(
                            'You don’t have read access to {font} or it doesn’t exists.'  # noqa
                        )
                        GLib.idle_add(
                            self._show_error, error_text.format(font=path)
                        )

            except Exception as exc:
                print(f'Error loading {path}')
                print(exc)
                error_text = _('Error: {error}.')
                GLib.idle_add(self._show_error, error_text.format(error=exc))

        self.window.processing = False

    def _show_error(self, text: str):
        error = Adw.Toast.new(text)
        self.window.toasts.add_toast(error)

    def _get_font_data(self, tt_font: TTFont) -> FontData:
        head = tt_font['head']
        naming = tt_font['name']  # Font naming table
        variations = None
        if 'fvar' in tt_font:
            variations = tt_font['fvar']

        font_data = FontData(
            naming.getBestFullName(),  # type: ignore
            naming.getBestFamilyName(),  # type: ignore
            naming.getDebugName(5),  # type: ignore
        )

        # Font macStyle bits
        mac_style_bitfield = [bool(int(b)) for b in bin(head.macStyle)[2:]]  # type: ignore
        mac_style_bitfield.reverse()

        # Get font bold or italic by macStyle
        if mac_style_bitfield[0]:
            font_data.weight = '700'  # Font is bold
        if len(mac_style_bitfield) > 1 and mac_style_bitfield[1]:
            font_data.style = 'italic'  # Font is italic

        # Variable font data
        if variations is not None:
            font_data.is_variable = True
            axes: dict[str, tuple[float, ...]] = variations.getAxes()  # type: ignore
            if wght := axes.get('wght'):
                font_data.weight = self.__get_font_weight_axis_range(wght)
            if wdth := axes.get('wdth'):
                font_data.weight = self.__get_font_width_axis_range(wdth)

        # Get local font name
        font_data.local.append('local("%s")' % naming.getDebugName(4))  # type: ignore
        if not naming.getDebugName(6) == naming.getDebugName(4):  # type: ignore
            font_data.local.append('local("%s")' % naming.getDebugName(6))  # type: ignore

        # Get style, weight and width
        # We try to guess this info from some name strings
        names = ' '.join(
            (
                naming.getDebugName(17) or naming.getDebugName(2),  # type: ignore
                naming.getDebugName(4),  # type: ignore
            )
        )
        ws = self.__normalize_names(names.split())

        for s in ws:
            if s == 'italic':
                font_data.style = 'italic'
            if variations is None:
                if s in WEIGHTS:
                    font_data.weight = WEIGHTS[s]
                elif s in WIDTHS:
                    font_data.width = WIDTHS[s]

        return font_data

    def __normalize_names(self, names: Sequence[str]) -> Sequence[str]:
        return list(
            map(lambda n: n.lower().replace('-', '').replace('_', ''), names)
        )

    def __get_font_weight_axis_range(self, weights: Sequence[float]) -> str:
        min_weight = min(weights)
        max_weight = max(weights)

        if min_weight == max_weight:
            weight = str(int(max_weight))
        else:
            weight = f"{int(min_weight)} {int(max_weight)}"

        return weight

    def __get_font_width_axis_range(self, widths: Sequence[float]) -> str:
        min_width = min(widths)
        max_width = max(widths)

        if min_width == max_width:
            width = f"{max_width:g}%"
        else:
            width = f"{min_width:g}% {max_width:g}%"

        return width



# Copyright 2020 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

from fontTools.subset import parse_unicodes
from gi.repository import Adw, Gio, Gtk

PRESET_RANGES = {
    'latin': 'U+0000-00FF,U+0131,U+0152-0153,U+02BB-02BC,U+02C6,U+02DA,U+02DC,'
    + 'U+2000-206F,U+2074,U+20AC,U+2122,U+2212,U+2215',
    'latin-ext': 'U+0100-024F,U+0259,U+1E00-1EFF,U+20A0-20CF,U+2C60-2C7F,'
    + 'U+A720-A7FF',
    'cyrillic': 'U+0400-045F,U+0490-0491,U+04B0-04B1,U+2116',
    'cyrillic-ext': 'U+0460-052F,U+1C80-1C88,U+20B4,U+2DE0-2DFF,U+A640-A69F,'
    + 'U+FE2E-FE2F',
    'greek': 'U+0370-03FF',
    'greek-ext': 'U+1F00-1FFF',
    'vietnamese': 'U+0102-0103,U+0110-0111,U+1EA0-1EF9,U+20AB',
    'devanagari': 'U+0900-097F,U+1CD0-1CF6,U+1CF8-1CF9,U+200B-200D,U+20A8,'
    + 'U+20B9,U+25CC,U+A830-A839,U+A8E0-A8FB',
}


@Gtk.Template(
    resource_path='/com/rafaelmardojai/WebfontKitGenerator/options.ui'
)
class Options(Adw.PreferencesPage):
    __gtype_name__ = 'Options'

    # Fonts format
    woff2: Gtk.Switch = Gtk.Template.Child()  # type: ignore
    woff: Gtk.Switch = Gtk.Template.Child()  # type: ignore

    # Subsetting
    subsetting: Adw.ExpanderRow = Gtk.Template.Child()  # type: ignore
    latin: Gtk.Switch = Gtk.Template.Child()  # type: ignore
    latin_ext: Gtk.Switch = Gtk.Template.Child()  # type: ignore
    cyrillic: Gtk.Switch = Gtk.Template.Child()  # type: ignore
    cyrillic_ext: Gtk.Switch = Gtk.Template.Child()  # type: ignore
    greek: Gtk.Switch = Gtk.Template.Child()  # type: ignore
    greek_ext: Gtk.Switch = Gtk.Template.Child()  # type: ignore
    vietnamese: Gtk.Switch = Gtk.Template.Child()  # type: ignore
    devanagari: Gtk.Switch = Gtk.Template.Child()  # type: ignore
    custom: Gtk.Entry = Gtk.Template.Child()  # type: ignore

    # CSS
    base64: Adw.SwitchRow = Gtk.Template.Child()  # type: ignore
    font_display: Adw.ComboRow = Gtk.Template.Child()  # type: ignore

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # GSettings object
        self.settings = Gio.Settings.new(
            'com.rafaelmardojai.WebfontKitGenerator'
        )

        self.load_saved()

    def load_saved(self):
        # Generate list of toggle names
        toggles = ['woff2', 'woff', 'base64']
        toggles.extend(PRESET_RANGES.keys())  # latin, latin_ext, etc

        for name in toggles:
            toggle = getattr(
                self, name.replace('-', '_')
            )  # Get Gtk.Switch name
            self.settings.bind(
                name, toggle, 'active', Gio.SettingsBindFlags.DEFAULT
            )  # Bind toggle with its corresponding GSetting

        self.settings.bind(
            'font-display',
            self.font_display,
            'selected',
            Gio.SettingsBindFlags.DEFAULT,
        )  # Bind font-display setting to combo row

        self.settings.bind(
            'subsetting',
            self.subsetting,
            'enable-expansion',
            Gio.SettingsBindFlags.DEFAULT,
        )  # Bind subsetting setting to expander row

    def get_formats(self) -> list[str]:
        """Get list of active formats"""
        formats = []

        if self.woff2.get_active():
            formats.append('woff2')
        if self.woff.get_active():
            formats.append('woff')

        if len(formats) == 0:
            formats = ['woff2']

        return formats

    def get_subsetting(self) -> dict[str, str] | None:
        """Get a dictionary with the active subsets and its unicode ranges"""

        if not self.subsetting.props.enable_expansion:
            return None

        ranges = {}

        # Add sunset if active
        for name in PRESET_RANGES.keys():
            toggle = getattr(self, name.replace('-', '_'))
            if toggle.get_active():
                ranges[name] = PRESET_RANGES[name]

        # Add custom subset
        try:
            if parse_unicodes(self.custom.get_text()):
                ranges['custom'] = self.custom.get_text()
        except Exception:
            pass

        if len(ranges) == 0:
            return None

        return ranges

    def get_base64(self) -> bool:
        return self.base64.props.active

    def get_font_display(self) -> str | None:
        if self.font_display.props.selected == 0:
            return None
        return (
            self.font_display.props.selected_item.props.string  # type: ignore
        )  # Get string from StringObject

    @Gtk.Template.Callback()
    def _validate_subsetting(self, _entry):
        try:
            parse_unicodes(self.custom.get_text())
            if self.custom.has_css_class('error'):
                self.custom.remove_css_class('error')
        except Exception:
            if not self.custom.has_css_class('error'):
                self.custom.add_css_class('error')



# Copyright 2020 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import os
from gettext import gettext as _
from typing import Sequence

from gi.repository import Adw, Gdk, Gio, GLib, GObject, Gtk

from webfontkitgenerator.font import Font, FontRow
from webfontkitgenerator.generator import Generator
from webfontkitgenerator.google import GoogleDialog
from webfontkitgenerator.loader import Loader
from webfontkitgenerator.log import Log
from webfontkitgenerator.options import Options
from webfontkitgenerator.sourceview import SourceView


@Gtk.Template(
    resource_path='/com/rafaelmardojai/WebfontKitGenerator/window.ui'
)
class Window(Adw.ApplicationWindow):
    __gtype_name__ = 'Window'

    processing: bool = GObject.Property(type=bool, default=False)  # type: ignore

    # Main app stack
    appstack: Gtk.Stack = Gtk.Template.Child()  # type: ignore

    # Progress View
    progressbar: Gtk.ProgressBar = Gtk.Template.Child()  # type: ignore
    progress: Adw.StatusPage = Gtk.Template.Child()  # type: ignore
    cancel: Gtk.Button = Gtk.Template.Child()  # type: ignore

    # Finished View
    finished_stack: Gtk.Stack = Gtk.Template.Child()  # type: ignore
    src_html: SourceView = Gtk.Template.Child()  # type: ignore
    src_css: SourceView = Gtk.Template.Child()  # type: ignore
    log: Log = Gtk.Template.Child()  # type: ignore

    # Fonts list and options view
    split: Adw.OverlaySplitView = Gtk.Template.Child()  # type: ignore
    toolbar_view: Adw.ToolbarView = Gtk.Template.Child()  # type: ignore
    fonts_stack: Gtk.Stack = Gtk.Template.Child()  # type: ignore
    fonts_list: Gtk.ListBox = Gtk.Template.Child()  # type: ignore
    options: Options = Gtk.Template.Child()  # type: ignore
    directory: Gtk.Label = Gtk.Template.Child()  # type: ignore
    toasts: Adw.ToastOverlay = Gtk.Template.Child()  # type: ignore

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Set default window icon for window managers
        self.set_default_icon_name('com.rafaelmardojai.WebfontKitGenerator')

        self.setup_actions()

        # Out directory chooser
        self.out_dir: Gio.File | None = None  # Store selected dir
        self.out_dir_chooser = Gtk.FileDialog()
        self.out_dir_chooser.props.title = _('Select Output Directory')

        # Fonts chooser
        self.fonts_chooser = Gtk.FileDialog()
        self.fonts_chooser.props.title = _('Open Font Files')

        fonts_filter = Gtk.FileFilter()
        fonts_filter.set_name(_('OTF & TTF'))
        fonts_filter.add_mime_type('font/otf')
        fonts_filter.add_pattern('.otf')
        fonts_filter.add_mime_type('font/ttf')
        fonts_filter.add_pattern('.ttf')

        filters = Gio.ListStore.new(Gtk.FileFilter)
        filters.append(fonts_filter)
        self.fonts_chooser.props.filters = filters

        # Setup fonts list
        self.model = Gio.ListStore.new(Font)
        self.model.connect(
            'items-changed', lambda _l, _p, _r, _a: self._check_ready_state()
        )
        self.fonts_list.bind_model(self.model, self._create_font_row)
        self.loader = Loader(self, self.model)  # Load fonts to the list model

        # Drag and drop
        drop_target = Gtk.DropTarget.new(Gdk.FileList, Gdk.DragAction.COPY)
        drop_target.connect('drop', self._on_drop)
        self.add_controller(drop_target)

    def setup_actions(self):
        remove_font = Gio.SimpleAction.new(
            'remove-font', GLib.VariantType.new('u')
        )
        remove_font.connect('activate', self._on_remove_font)
        self.add_action(remove_font)

        set_dir = Gio.SimpleAction.new('set-dir', None)
        set_dir.connect('activate', self._on_set_dir)
        self.add_action(set_dir)

        open_ = Gio.SimpleAction.new('open', None)
        open_.connect('activate', self._on_open)
        self.add_action(open_)

        google = Gio.SimpleAction.new('google', None)
        google.connect('activate', self._on_google)
        self.add_action(google)

        generate = Gio.SimpleAction.new('generate', None)
        generate.connect('activate', self._on_generate)
        generate.set_enabled(False)
        self.add_action(generate)

        back = Gio.SimpleAction.new('back', None)
        back.connect('activate', self._on_back)
        self.add_action(back)

    def load_fonts(self, files: Sequence[Gio.File]):
        self.loader.load(files)

    @Gtk.Template.Callback()
    def _open_generation_dir(self, _button):
        def on_launched(launcher, result):
            launcher.launch_finish(result)

        launcher = Gtk.FileLauncher(file=self.out_dir)
        launcher.launch(self, None, on_launched)

    @Gtk.Template.Callback()
    def _on_appstack_changes(self, stack: Gtk.Stack, _pspec):
        file_choosers = stack.props.visible_child_name == 'main'

        self.lookup_action('open').set_enabled(file_choosers)  # type: ignore
        self.lookup_action('set-dir').set_enabled(file_choosers)  # type: ignore

    def _on_open(self, _action, _param):
        def on_selected(chooser: Gtk.FileDialog, result: Gio.AsyncResult):
            try:
                files = chooser.open_multiple_finish(result)

                if files is not None:
                    self.load_fonts(files)  # type: ignore

            except GLib.Error as e:
                if e.code == Gtk.DialogError.FAILED:
                    print(e.code)

        self.fonts_chooser.open_multiple(self, None, on_selected)

    def _on_set_dir(self, _action, _param):
        def on_selected(chooser: Gtk.FileDialog, result: Gio.AsyncResult):
            try:
                selected = chooser.select_folder_finish(result)

                if selected is not None and (path := selected.get_path()):
                    if os.access(path, os.W_OK):
                        self.out_dir = selected
                        self.directory.set_label(selected.get_basename() or "")
                        self._check_ready_state()
                    else:
                        self.toasts.add_toast(
                            Adw.Toast.new(
                                _(
                                    'You don’t have write access to the selected directory.'  # noqa
                                )
                            )
                        )
            except GLib.Error as e:
                if e.code == Gtk.DialogError.FAILED:
                    print(e.code)

        self.out_dir_chooser.select_folder(self, None, on_selected)

    def _on_google(self, _action, _param):
        dialog = GoogleDialog(self)
        dialog.present(self)

    def _on_generate(self, _action, _param):
        if self.out_dir and (path := self.out_dir.get_path()):
            generator = Generator(
                self,
                path,
                self.model,  # type:ignore
                self.options.get_formats(),
                self.options.get_subsetting(),
                self.options.get_base64(),
                self.options.get_font_display(),
            )
            generator.run()

    def _on_back(self, _action, _param):
        if not self.processing:
            self.appstack.set_visible_child_name('main')

    def _on_remove_font(self, _action, param):
        self.model.remove(param.get_uint32())

    def _create_font_row(self, font: Font):
        widget = FontRow(font.data)
        return widget

    def _on_drop(self, _target, value, _x, _y):
        if (
            isinstance(value, Gdk.FileList)
            and self.appstack.get_visible_child_name() == 'main'
        ):
            self.load_fonts(value.get_files())

    def _check_ready_state(self):
        has_items = self.model.get_n_items() > 0

        self.lookup_action('generate').set_enabled(has_items and self.out_dir)  # type: ignore
        self.split.props.show_sidebar = has_items
        self.toolbar_view.props.reveal_bottom_bars = has_items

        if has_items:
            self.fonts_stack.set_visible_child_name('fonts')
            self.toolbar_view.props.top_bar_style = Adw.ToolbarStyle.RAISED
        else:
            self.fonts_stack.set_visible_child_name('empty')
            self.toolbar_view.props.top_bar_style = Adw.ToolbarStyle.FLAT
